﻿using System;

namespace NumberTypes
{
    /// <summary>
    /// Racionális szám típusa
    /// </summary>
    class Rational : IRealCompatible
    {
        private Int32 _Nominator;
        private Int32 _Denominator;

        /// <summary>
        /// Számláló
        /// </summary>
        public Int32 Nominator
        {
            get { return _Nominator; }
            set { _Nominator = value; }
        }

        /// <summary>
        /// Nevező
        /// </summary>
        public Int32 Denominator
        {
            get { return _Denominator; }
            set { _Denominator = (value != 0) ? value : 1; }
        }

        /// <summary>
        /// Racionális szám létrehozása
        /// </summary>
        public Rational()
        {
            Nominator = 0;
            Denominator = 1;
        }

        /// <summary>
        /// Racionális szám létrehozása
        /// </summary>
        /// <param name="nominator">Számláló</param>
        /// <param name="denominator">Nevező</param>
        public Rational(Int32 nominator, Int32 denominator)
        {
            Nominator = nominator;
            Denominator = denominator;
            Simplify(ref _Nominator, ref _Denominator);
        }

        /// <summary>
        /// Racionális számok összeadása
        /// </summary>
        /// <param name="first">Első operandus</param>
        /// <param name="second">Második operandus</param>
        /// <returns>Összeadás eredménye</returns>
        public static Rational operator +(Rational first, Rational second)
        {
            Rational result = new Rational();
            result.Nominator = first.Nominator * second.Denominator + first.Denominator * second.Nominator;
            result.Denominator = first.Denominator * second.Denominator;
            Simplify(ref result._Nominator, ref result._Denominator);
            return result;
        }

        /// <summary>
        /// Racionális számok különbsége
        /// </summary>
        /// <param name="first">Első operandus</param>
        /// <param name="second">Második operandus</param>
        /// <returns>Kivonás eredménye</returns>
        public static Rational operator -(Rational first, Rational second)
        {
            Rational result = new Rational();
            result.Nominator = first.Nominator * second.Denominator - first.Denominator * second.Nominator;
            result.Denominator = first.Denominator * second.Denominator;
            Simplify(ref result._Nominator, ref result._Denominator);
            return result;
        }

        /// <summary>
        /// Racionális számok szorzása
        /// </summary>
        /// <param name="first">Első operandus</param>
        /// <param name="second">Második operandus</param>
        /// <returns>Szorzás eredménye</returns>
        public static Rational operator *(Rational first, Rational second)
        {
            Rational result = new Rational();
            result.Nominator = first.Nominator * second.Nominator;
            result.Denominator = first.Denominator * second.Denominator;
            Simplify(ref result._Nominator, ref result._Denominator);
            return result;
        }

        /// <summary>
        /// Racionális számok hányadosa
        /// </summary>
        /// <param name="first">Első operandus</param>
        /// <param name="second">Második operandus</param>
        /// <returns>Osztás eredménye</returns>
        public static Rational operator /(Rational first, Rational second)
        {
            Rational result = new Rational();
            result.Nominator = first.Nominator * second.Denominator;
            result.Denominator = first.Denominator * second.Nominator;
            Simplify(ref result._Nominator, ref result._Denominator);
            return result;
        }

        /// <summary>
        /// Egyszerűsítés művelete
        /// </summary>
        /// <param name="first">Első érték</param>
        /// <param name="second">Második érték</param>
        private static void Simplify(ref Int32 first, ref Int32 second)
        {
            Int32 m = first, n = second;
            if (m < 1 || n < 1)
                return;

            // Euklideszi algoritmus futtatása
            while (m != n)
            {
                if (m > n)
                    m -= n;
                else
                    n -= m;
            }

            // Egyszerűsítés a legnagyobb közös osztóval
            first /= m;
            second /= m;
        }

        /// <summary>
        /// Lebegőpontos érték
        /// </summary>
        public Double Value
        {
            get { return (Double)Nominator / Denominator; }
        }

        /// <summary>
        /// Szöveggé alakítás
        /// </summary>
        /// <returns>Racionális szám szövegszerű leírása</returns>
        public override String ToString()
        {
            return String.Format("{0} / {1}", Nominator, Denominator);
        }

        #region IRealCompatible Members

        public Double ToDouble()
        {
            return (Double)_Nominator / _Denominator;
        }

        public void FromDouble(Double val)
        {
            _Nominator = Convert.ToInt32(Math.Round(val, 5, MidpointRounding.ToEven) * 10000);
            _Denominator = 10000;
            Simplify(ref _Nominator, ref _Denominator);
        }

        #endregion
    }
}
